@@ -13,6 +13,7 @@ class RESTMessageResponse(object):
     """
     A base protobuf wrapper class, all api response use.
     """
+
     def __init__(self, rest_data):
         self._rest_data = rest_data
 
@@ -34,6 +35,11 @@ class RESTMessageResponse(object):
     def data_v1(self):
         return self._rest_data
 
+    @classmethod
+    def _unix_epoch_field_value(cls, unix_epoch_val):
+        dt = datetime.fromtimestamp(unix_epoch_val / 1000)
+        return dt.replace(microsecond=(unix_epoch_val % 1000) * 1000)
+
     def __nonzero__(self):
         return self.__bool__()
 
@@ -52,6 +58,7 @@ class ApiCallResponse(RESTMessageResponse):
     This is a wrapper class for a proto MsgApiCallResponse.
     It provides some additional methods for easier state checking of the ApiCallResponse.
     """
+
     def __init__(self, rest_data):
         super(ApiCallResponse, self).__init__(rest_data)
 
@@ -65,6 +72,10 @@ class ApiCallResponse(RESTMessageResponse):
         """
         return ApiCallResponse(json_data)
 
+    @classmethod
+    def from_str(cls, message, code=None):
+        return ApiCallResponse({"message": message, "ret_code": apiconsts.FAIL_UNKNOWN_ERROR})
+
     def is_error(self, code=None):
         """
         Returns True if the ApiCallResponse is any error and "code" is unset.
@@ -337,6 +348,10 @@ class Node(RESTMessageResponse):
     def flags(self):
         return self._rest_data.get("flags", [])
 
+    @property
+    def eviction_timestamp(self):
+        return self._rest_data.get("eviction_timestamp")
+
     @property
     def data_v0(self):
         d = dict(self._rest_data)
@@ -384,10 +399,61 @@ class NodeListResponse(RESTMessageResponse):
         }
 
 
+class NodeConnection(RESTMessageResponse):
+    def __init__(self, data):
+        super(NodeConnection, self).__init__(data)
+
+    @property
+    def node_a(self):
+        return self._rest_data["node_a"]
+
+    @property
+    def node_b(self):
+        return self._rest_data["node_b"]
+
+    @property
+    def flags(self):
+        return self._rest_data.get("flags", [])
+
+    @property
+    def properties(self):
+        return self._rest_data.get("props", {})
+
+    @property
+    def data_v0(self):
+        d = {
+            "node_name_1": self.node_a,
+            "node_name_2": self.node_b
+        }
+        if self.flags:
+            d["flags"] = self.flags
+        if self.properties:
+            print("self properties:")
+            print(self.properties)
+            d["props"] = [{"key": x, "value": v} for x, v in self.properties.items()]
+        return d
+
+
+class NodeConnectionsResponse(RESTMessageResponse):
+    def __init__(self, data):
+        super(NodeConnectionsResponse, self).__init__(data)
+
+    @property
+    def node_connections(self):
+        return [NodeConnection(x) for x in self._rest_data]
+
+    @property
+    def data_v0(self):
+        d = {}
+        if self.node_connections:
+            d["node_connections"] = [x.data_v0 for x in self.node_connections]
+        return d
+
+
 class FreeSpace(RESTMessageResponse):
     def __init__(self, rest_data):
         super(FreeSpace, self).__init__(rest_data)
-        
+
     @property
     def free_capacity(self):
         return self._rest_data.get("free_capacity")
@@ -397,7 +463,8 @@ class FreeSpace(RESTMessageResponse):
         return self._rest_data.get("total_capacity")
 
     def __str__(self):
-        return "{used}/{total} Kib used".format(used=self.total_capacity-self.free_capacity, total=self.total_capacity)
+        return "{used}/{total} Kib used".format(
+            used=self.total_capacity - self.free_capacity, total=self.total_capacity)
 
     @property
     def data_v0(self):
@@ -420,6 +487,10 @@ class StoragePoolDriver(object):
     REMOTE_SPDK = "REMOTE_SPDK"
     OPENFLEX_TARGET = "OPENFLEX_TARGET"
     EXOS = "EXOS"
+    STORAGE_SPACES = "STORAGE_SPACES"
+    STORAGE_SPACES_THIN = "STORAGE_SPACES_THIN"
+    EBS_TARGET = "EBS_TARGET"
+    EBS_INIT = "EBS_INIT"
 
     @staticmethod
     def list():
@@ -434,13 +505,18 @@ class StoragePoolDriver(object):
             StoragePoolDriver.SPDK,
             StoragePoolDriver.REMOTE_SPDK,
             StoragePoolDriver.OPENFLEX_TARGET,
-            StoragePoolDriver.EXOS
+            StoragePoolDriver.EXOS,
+            StoragePoolDriver.STORAGE_SPACES,
+            StoragePoolDriver.STORAGE_SPACES_THIN,
+            StoragePoolDriver.EBS_TARGET,
+            StoragePoolDriver.EBS_INIT
         ]
 
     @classmethod
     def diskless_driver(cls):
         return [
-            StoragePoolDriver.Diskless
+            StoragePoolDriver.Diskless,
+            StoragePoolDriver.EBS_INIT
         ]
 
     @staticmethod
@@ -484,6 +560,16 @@ class StoragePoolDriver(object):
                 apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_OPENFLEX_STOR_POOL: driver_pool_name
             }
 
+        if storage_driver == StoragePoolDriver.STORAGE_SPACES:
+            return {
+                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_STORAGE_SPACES: driver_pool_name
+            }
+
+        if storage_driver == StoragePoolDriver.STORAGE_SPACES_THIN:
+            return {
+                apiconsts.NAMESPC_STORAGE_DRIVER + '/' + apiconsts.KEY_STOR_POOL_STORAGE_SPACES: driver_pool_name
+            }
+
         raise LinstorError(
             "Unknown storage driver '{drv}', known drivers: "
             "lvm, lvmthin, zfs, diskless, spdk".format(drv=storage_driver)
@@ -1590,7 +1676,7 @@ class Resource(RESTMessageResponse):
         :rtype: Optional[datetime]
         """
         if "create_timestamp" in self._rest_data:
-            return datetime.fromtimestamp(self._rest_data["create_timestamp"]/1000)
+            return datetime.fromtimestamp(self._rest_data["create_timestamp"] / 1000)
         return None
 
     @property
@@ -1762,13 +1848,38 @@ class Snapshot(RESTMessageResponse):
         :rtype: Optional[datetime]
         """
         if "create_timestamp" in self._rest_data:
-            return datetime.fromtimestamp(self._rest_data["create_timestamp"]/1000)
+            return datetime.fromtimestamp(self._rest_data["create_timestamp"] / 1000)
         return None
 
     @property
     def uuid(self):
         return self._rest_data.get("uuid")
 
+    @property
+    def snapshot_volumes(self):
+        return [SnapshotVolume(x) for x in self._rest_data.get("snapshot_volumes", [])]
+
+
+class SnapshotVolume(RESTMessageResponse):
+    def __init__(self, data):
+        super(SnapshotVolume, self).__init__(data)
+
+    @property
+    def uuid(self):
+        return self._rest_data.get("uuid")
+
+    @property
+    def vlm_nr(self):
+        return self._rest_data.get("vlm_nr")
+
+    @property
+    def state(self):
+        return self._rest_data.get("state")
+
+    @property
+    def properties(self):
+        return self._rest_data.get("props", {})
+
 
 class SnapshotDefinition(RESTMessageResponse):
     def __init__(self, data):
@@ -2235,3 +2346,677 @@ class ExosMapListResponse(RESTMessageResponse):
         :rtype: list[ExosMapResponse]
         """
         return [ExosMapResponse(x) for x in self._rest_data]
+
+
+class CloneStarted(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(CloneStarted, self).__init__(rest_data)
+
+    @property
+    def location(self):
+        """
+        API path for further status updates.
+        :return: e.g. "/v1/resource-definitions/testrsc/clone/clonersc"
+        :rtype: str
+        """
+        return self._rest_data["location"]
+
+    @property
+    def source_name(self):
+        """
+        Name of the source resource.
+
+        :return:
+        :rtype: str
+        """
+        return self._rest_data["source_name"]
+
+    @property
+    def clone_name(self):
+        """
+        Name of the cloned resource.
+
+        :return:
+        :rtype: str
+        """
+        return self._rest_data["clone_name"]
+
+    @property
+    def messages(self):
+        """
+
+        :return:
+        :rtype: list[ApiCallResponse]
+        """
+        return [ApiCallResponse(x) for x in self._rest_data.get("messages", [])]
+
+
+class CloneStatus(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(CloneStatus, self).__init__(rest_data)
+
+    @property
+    def status(self):
+        """
+        Current status of the clone operation.
+
+        :return:
+        :rtype: apiconsts.CloneStatus
+        """
+        return apiconsts.CloneStatus(self._rest_data["status"])
+
+
+class SyncStatus(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(SyncStatus, self).__init__(rest_data)
+
+    @property
+    def synced_on_all(self):
+        """
+        Current sync status of the resource definition.
+
+        :return: True if resource is ready to be e.g. resized
+        :rtype: bool
+        """
+        return self._rest_data["synced_on_all"]
+
+
+class S3Remote(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(S3Remote, self).__init__(rest_data)
+
+    @property
+    def remote_name(self):
+        return self._rest_data["remote_name"]
+
+    @property
+    def endpoint(self):
+        return self._rest_data["endpoint"]
+
+    @property
+    def bucket(self):
+        return self._rest_data["bucket"]
+
+    @property
+    def region(self):
+        return self._rest_data["region"]
+
+    @property
+    def access_key(self):
+        return self._rest_data.get("access_key")
+
+    @property
+    def secret_key(self):
+        return self._rest_data.get("secret_key")
+
+
+class LinstorRemote(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(LinstorRemote, self).__init__(rest_data)
+
+    @property
+    def remote_name(self):
+        return self._rest_data["remote_name"]
+
+    @property
+    def url(self):
+        return self._rest_data["url"]
+
+    @property
+    def passphrase(self):
+        return self._rest_data.get("passphrase")
+
+
+class EbsRemote(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(EbsRemote, self).__init__(rest_data)
+
+    @property
+    def remote_name(self):
+        return self._rest_data["remote_name"]
+
+    @property
+    def endpoint(self):
+        return self._rest_data["endpoint"]
+
+    @property
+    def region(self):
+        return self._rest_data.get("region")
+
+    @property
+    def availability_zone(self):
+        return self._rest_data.get("availability_zone")
+
+
+class RemoteListResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(RemoteListResponse, self).__init__(rest_data)
+
+    @property
+    def s3_remotes(self):
+        """
+
+        :return:
+        :rtype: list[S3Remote]
+        """
+        return [S3Remote(x) for x in self._rest_data.get("s3_remotes", [])]
+
+    @property
+    def linstor_remotes(self):
+        """
+
+        :return:
+        :rtype: list[LinstorRemote]
+        """
+        return [LinstorRemote(x) for x in self._rest_data.get("linstor_remotes", [])]
+
+    @property
+    def ebs_remotes(self):
+        """
+
+        :return:
+        :rtype: list[EbsRemote]
+        """
+        return [EbsRemote(x) for x in self._rest_data.get("ebs_remotes", [])]
+
+
+class Backup(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(Backup, self).__init__(rest_data)
+
+    @property
+    def id(self):
+        return self._rest_data.get("id")
+
+    @property
+    def start_time(self):
+        return self._rest_data.get("start_time")
+
+    @property
+    def start_timestamp(self):
+        return self._rest_data.get("start_timestamp")
+
+    @property
+    def finished_time(self):
+        return self._rest_data.get("finished_time")
+
+    @property
+    def finished_timestamp(self):
+        return self._rest_data.get("finished_timestamp")
+
+    @property
+    def origin_rsc_name(self):
+        return self._rest_data.get("origin_rsc")
+
+    @property
+    def origin_snap_name(self):
+        return self._rest_data.get("origin_snap")
+
+    @property
+    def based_on(self):
+        return self._rest_data.get("based_on_id")
+
+    @property
+    def success(self):
+        return self._rest_data.get("success")
+
+    @property
+    def shipping(self):
+        return self._rest_data.get("shipping")
+
+    @property
+    def restorable(self):
+        return self._rest_data.get("restorable")
+
+
+class BackupOther(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(BackupOther, self).__init__(rest_data)
+
+    @property
+    def files(self):
+        """
+        Backup file list
+        :return:
+        :rtype: list[str]
+        """
+        return self._rest_data.get("files", [])
+
+
+class BackupListResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(BackupListResponse, self).__init__(rest_data)
+
+    @property
+    def linstor(self):
+        """
+        :return:
+        :rtype: list[Backup]
+        """
+        return [Backup(v) for k, v in self._rest_data.get("linstor", {}).items()]
+
+    @property
+    def other(self):
+        """
+        :return:
+        :rtype: list[BackupOther]
+        """
+        return BackupOther(self._rest_data.get("other", {}))
+
+
+class BackupInfoResponse(RESTMessageResponse):
+
+    def __init__(self, rest_data):
+        super(BackupInfoResponse, self).__init__(rest_data)
+
+    @property
+    def rsc(self):
+        return self._rest_data.get("rsc")
+
+    @property
+    def full(self):
+        return self._rest_data.get("full")
+
+    @property
+    def snap(self):
+        return self._rest_data.get("snap")
+
+    @property
+    def latest(self):
+        return self._rest_data.get("latest")
+
+    @property
+    def count(self):
+        return self._rest_data.get("count")
+
+    @property
+    def dl_size(self):
+        return self._rest_data.get("dl_size_kib")
+
+    @property
+    def alloc_size(self):
+        return self._rest_data.get("alloc_size_kib")
+
+    @property
+    def storpools(self):
+        """
+        :return:
+        :rtype: list[BackupInfoStorPool]
+        """
+        return [BackupInfoStorPool(v) for v in self._rest_data.get("storpools", [])]
+
+
+class BackupInfoStorPool(RESTMessageResponse):
+
+    def __init__(self, rest_data):
+        super(BackupInfoStorPool, self).__init__(rest_data)
+
+    @property
+    def name(self):
+        return self._rest_data.get("name")
+
+    @property
+    def provider_kind(self):
+        return self._rest_data.get("provider_kind")
+
+    @property
+    def target_name(self):
+        return self._rest_data.get("target_name")
+
+    @property
+    def remaining_space(self):
+        return self._rest_data.get("remaining_space_kib")
+
+    @property
+    def volumes(self):
+        """
+        :return:
+        :rtype: list[BackupInfoStorPoolVolume]
+        """
+        return [BackupInfoStorPoolVolume(v) for v in self._rest_data.get("vlms", [])]
+
+
+class BackupInfoStorPoolVolume(RESTMessageResponse):
+
+    def __init__(self, rest_data):
+        super(BackupInfoStorPoolVolume, self).__init__(rest_data)
+
+    @property
+    def name(self):
+        return self._rest_data.get("name")
+
+    @property
+    def layer_type(self):
+        return self._rest_data.get("layer_type")
+
+    @property
+    def dl_size(self):
+        return self._rest_data.get("dl_size_kib")
+
+    @property
+    def alloc_size(self):
+        return self._rest_data.get("alloc_size_kib")
+
+    @property
+    def usable_size(self):
+        return self._rest_data.get("usable_size_kib")
+
+
+class BackupQueues(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(BackupQueues, self).__init__(rest_data)
+
+    @property
+    def node_queues(self):
+        """
+        :rtype: list[NodeQueue]
+        """
+        return [NodeQueue(v) for v in self._rest_data.get("node_queues", [])]
+
+    @property
+    def snap_queues(self):
+        """
+        :rtype: list[SnapQueue]
+        """
+        return [SnapQueue(v) for v in self._rest_data.get("snap_queues", [])]
+
+
+class NodeQueue(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(NodeQueue, self).__init__(rest_data)
+
+    @property
+    def node_name(self):
+        return self._rest_data.get("node_name")
+
+    @property
+    def queue(self):
+        """
+        :rtype: list[SnapQueue]
+        """
+        return [SnapQueue(v) for v in self._rest_data.get("queue", [])]
+
+
+class SnapQueue(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(SnapQueue, self).__init__(rest_data)
+
+    @property
+    def resource_name(self):
+        return self._rest_data.get("resource_name")
+
+    @property
+    def snapshot_name(self):
+        return self._rest_data.get("snapshot_name")
+
+    @property
+    def remote_name(self):
+        return self._rest_data.get("remote_name")
+
+    @property
+    def incremental(self):
+        return self._rest_data.get("incremental")
+
+    @property
+    def based_on(self):
+        return self._rest_data.get("based_on")
+
+    @property
+    def start_timestamp(self):
+        return self._rest_data.get("start_timestamp")
+
+    @property
+    def pref_node(self):
+        return self._rest_data.get("pref_node")
+
+    @property
+    def queue(self):
+        """
+        :rtype: list[NodeQueue]
+        """
+        return [NodeQueue(v) for v in self._rest_data.get("queue", [])]
+
+
+class ExternalFile(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(ExternalFile, self).__init__(rest_data)
+
+    @property
+    def path(self):
+        return self._rest_data.get("path")
+
+    @property
+    def content(self):
+        return self._rest_data.get("content")
+
+
+class FileResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(FileResponse, self).__init__(rest_data)
+
+    @property
+    def files(self):
+        if isinstance(self._rest_data, list):
+            return [ExternalFile(x) for x in self._rest_data]
+        return ExternalFile(self._rest_data)
+
+
+class Schedule(RESTMessageResponse):
+    def __int__(self, rest_data):
+        super(Schedule, self).__init__(rest_data)
+
+    @property
+    def schedule_name(self):
+        return self._rest_data["schedule_name"]
+
+    @property
+    def full_cron(self):
+        return self._rest_data["full_cron"]
+
+    @property
+    def inc_cron(self):
+        return self._rest_data.get("inc_cron")
+
+    @property
+    def keep_local(self):
+        return self._rest_data.get("keep_local")
+
+    @property
+    def keep_remote(self):
+        return self._rest_data.get("keep_remote")
+
+    @property
+    def on_failure(self):
+        return self._rest_data.get("on_failure")
+
+    @property
+    def max_retries(self):
+        return self._rest_data.get("max_retries")
+
+
+class ScheduleListResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(ScheduleListResponse, self).__init__(rest_data)
+
+    @property
+    def schedules(self):
+        return [Schedule(x) for x in self._rest_data.get("data", [])]
+
+
+class ScheduleResource(RESTMessageResponse):
+    def __int__(self, rest_data):
+        super(ScheduleResource, self).__init__(rest_data)
+
+    @property
+    def resource_name(self):
+        return self._rest_data["rsc_name"]
+
+    @property
+    def remote_name(self):
+        return self._rest_data.get("remote_name")
+
+    @property
+    def schedule_name(self):
+        return self._rest_data.get("schedule_name")
+
+    @property
+    def reason(self):
+        return self._rest_data.get("reason")
+
+    @property
+    def last_snap_time(self):
+        """
+
+        :return:
+        :rtype: datetime
+        """
+        ts = self._rest_data.get("last_snap_time", -1)
+        return None if ts < 0 else self._unix_epoch_field_value(ts)
+
+    @property
+    def last_snap_inc(self):
+        """
+
+        :return:
+        :rtype: bool
+        """
+        return self._rest_data.get("last_snap_inc", False)
+
+    @property
+    def next_exec_time(self):
+        """
+
+        :return:
+        :rtype: datetime
+        """
+        ts = self._rest_data.get("next_exec_time", -1)
+        return None if ts < 0 else self._unix_epoch_field_value(ts)
+
+    @property
+    def next_exec_inc(self):
+        """
+
+        :return:
+        :rtype: bool
+        """
+        return self._rest_data.get("next_exec_inc", False)
+
+    @property
+    def next_planned_full(self):
+        """
+
+        :return:
+        :rtype: datetime
+        """
+        ts = self._rest_data.get("next_planned_full", -1)
+        return None if ts < 0 else self._unix_epoch_field_value(ts)
+
+    @property
+    def next_planned_inc(self):
+        """
+
+        :return:
+        :rtype: datetime
+        """
+        ts = self._rest_data.get("next_planned_inc", -1)
+        return None if ts < 0 else self._unix_epoch_field_value(ts)
+
+
+class ScheduleResourceListResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(ScheduleResourceListResponse, self).__init__(rest_data)
+
+    @property
+    def schedule_resources(self):
+        return [ScheduleResource(x) for x in self._rest_data.get("data", [])]
+
+
+class ScheduleResourceDetails(RESTMessageResponse):
+    def __int__(self, rest_data):
+        super(ScheduleResourceDetails, self).__init__(rest_data)
+
+    @property
+    def remote_name(self):
+        return self._rest_data["remote_name"]
+
+    @property
+    def schedule_name(self):
+        return self._rest_data["schedule_name"]
+
+    @property
+    def controller(self):
+        return self._rest_data.get("ctrl")
+
+    @property
+    def resource_definition(self):
+        return self._rest_data.get("rsc_dfn")
+
+    @property
+    def resource_group(self):
+        return self._rest_data.get("rsc_grp")
+
+
+class ScheduleResourceDetailsListResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(ScheduleResourceDetailsListResponse, self).__init__(rest_data)
+
+    @property
+    def schedule_resources(self):
+        return [ScheduleResourceDetails(x) for x in self._rest_data.get("data", [])]
+
+
+class QuerySizeInfoSpawnResult(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(QuerySizeInfoSpawnResult, self).__init__(rest_data)
+
+    @property
+    def node_name(self):
+        return self._rest_data["node_name"]
+
+    @property
+    def stor_pool_name(self):
+        return self._rest_data["stor_pool_name"]
+
+    @property
+    def stor_pool_oversubscription_ratio(self):
+        return self._rest_data["stor_pool_oversubscription_ratio"]
+
+
+class QuerySizeInfoResponseSpaceInfo(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(QuerySizeInfoResponseSpaceInfo, self).__init__(rest_data)
+
+    @property
+    def max_vlm_size_in_kib(self):
+        return self._rest_data["max_vlm_size_in_kib"]
+
+    @property
+    def available_size_in_kib(self):
+        return int(self._rest_data["available_size_in_kib"]) if "available_size_in_kib" in self._rest_data else None
+
+    @property
+    def capacity_in_kib(self):
+        return int(self._rest_data["capacity_in_kib"]) if "capacity_in_kib" in self._rest_data else None
+
+    @property
+    def default_max_oversubscription_ratio(self):
+        return float(self._rest_data["default_max_oversubscription_ratio"])\
+            if "default_max_oversubscription_ratio" in self._rest_data else None
+
+    @property
+    def next_spawn_result(self):
+        return [QuerySizeInfoSpawnResult(x) for x in self._rest_data["next_spawn_result"]]\
+            if "next_spawn_result" in self._rest_data else None
+
+
+class QuerySizeInfoResponse(RESTMessageResponse):
+    def __init__(self, rest_data):
+        super(QuerySizeInfoResponse, self).__init__(rest_data)
+
+    @property
+    def space_info(self):
+        return QuerySizeInfoResponseSpaceInfo(self._rest_data["space_info"])
+
+    @property
+    def reports(self):
+        return [ApiCallResponse(x) for x in self._rest_data.get("reports", [])]
